import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.checkpoint
from torch.nn import TransformerEncoderLayer
from transformers import AutoModel
from utils.loss import CrossEntropyLoss, KLDivergenceLoss, MMD, InfoNCE_KDLoss, ContrastiveLoss
from transformers import Wav2Vec2Model, WhisperModel, HubertModel
from transformers import WavLMModel
from modules.transformer import TransformerEncoderLayer as Cross_model_encoder
from modules.adapter import BottleneckAdapter as Adapter
from modules.ConNet import ConNet
from utils.helper import PredcitLayer
from .model import MyModel


class DQ_old(MyModel):
    def __init__(self,args):
        super(DQ_old, self).__init__(args)
        self.adapter_mode = ""
        self.frozen_mode = ""
        assert self.adapter_mode in ['t', 'a', 'ta', 'tv', 'av', 'tav', 'v', ''], "Unknown adapter mode"
        assert self.frozen_mode in ['t', 'a', 'ta', 'tv', 'av', 'tav', 'v', ''], "Unknown frozen mode"

        self.num_heads = args.num_heads
        self.fusion_layers = args.fusion_layers
        self.relu_dropout = args.relu_dropout
        self.res_dropout = args.res_dropout
        self.adapter_dim = args.adapter_dim
        self.bottleneck_len = args.bottleneck_len
        self.tau = args.tau
        self.t_loss_weight = args.t_loss_weight
        self.a_loss_weight = args.a_loss_weight
        self.f_loss_weight = args.f_loss_weight
    
        self.text_encoder = self.get_text_encoder(self.adapter_mode, self.frozen_mode)
        self.audio_encoder = self.get_audio_encoder(self.adapter_mode, self.frozen_mode)
        self.bottleneck_embedding = nn.Embedding(self.bottleneck_len, self.intermediate_dim)
        # self.bottleneck = torch.LongTensor([i for i in range(self.bottleneck_len)]).to("cuda:0")
        # self.bottleneck = nn.Parameter(torch.zeros(1, self.bottleneck_len, self.intermediate_dim, requires_grad=True)).to("cuda")
        # self.b_t_layers = nn.ModuleList([
        #     TransformerEncoderLayer(d_model=self.intermediate_dim, nhead=self.num_heads, batch_first=True, dropout=self.dropout, activation='gelu')
        #     for _ in range(self.fusion_layers)])
        # self.b_a_layers = nn.ModuleList([
        #     TransformerEncoderLayer(d_model=self.intermediate_dim, nhead=self.num_heads, batch_first=True, dropout=self.dropout, activation='gelu')
        #     for _ in range(self.fusion_layers)])
        
        self.query_encoder = nn.ModuleList([
            TransformerEncoderLayer(d_model=self.intermediate_dim, nhead=self.num_heads, batch_first=False, dropout=self.dropout, activation='gelu')
            for _ in range(self.fusion_layers)])
        # cross modal encoder
        self.cme_qt = nn.ModuleList([
            Cross_model_encoder(embed_dim=self.intermediate_dim, num_heads=self.num_heads, attn_dropout=self.dropout, relu_dropout=self.relu_dropout, res_dropout=self.res_dropout)
            for _ in range(self.fusion_layers)])
        self.cme_qa = nn.ModuleList([
            Cross_model_encoder(embed_dim=self.intermediate_dim, num_heads=self.num_heads, attn_dropout=self.dropout, relu_dropout=self.relu_dropout, res_dropout=self.res_dropout)
            for _ in range(self.fusion_layers)])
        # 0404 add
        
        # add unimodal fusion weight to adaptively fuse the unimodal representations 
        self.conf_net = ConNet(self.intermediate_dim)
        self.a_weight = nn.Parameter(torch.tensor(0.5, requires_grad=True)).to("cuda")
        self.t_weight = 1-self.a_weight
        # self.conf_nets = nn.ModuleList([ConNet(self.intermediate_dim) for _ in range(self.fusion_layers)])
    
        # prediciton head
        self.pred_head_t = PredcitLayer(self.intermediate_dim, self.output_dim, dropout=self.dropout, residual=True)
        self.pred_head_a = PredcitLayer(self.intermediate_dim, self.output_dim, dropout=self.dropout, residual=True)
        self.pred_head_union = PredcitLayer(self.intermediate_dim, self.output_dim, dropout=self.dropout, residual=True)


    def get_text_encoder(self, adapter_mode, frozen_mode):
        text_encoder = AutoModel.from_pretrained(self.args.text_model)

        if 't' in frozen_mode:
            # freeze all the parameters of the model
            for param in text_encoder.parameters():
                param.requires_grad = False
        elif 't' in adapter_mode:
            for i in range(0, text_encoder.config.num_hidden_layers):
                hidden_size = text_encoder.config.hidden_size
                adapter = Adapter(hidden_size, hidden_size//4)
                setattr(text_encoder.encoder.layer[i].attention.output.dense, 'adapter', adapter)
                setattr(text_encoder.encoder.layer[i].output.dense, 'adapter', adapter)

            for param_name, param in text_encoder.named_parameters():
                # can only tune the adapter parameters and LayerNorm parameters 
                if 'adapter' in param_name or 'LayerNorm' in param_name:
                    param.requires_grad = True
                else:
                    param.requires_grad = False
        
        return text_encoder


    def get_audio_encoder(self, adapter_mode, frozen_mode):
        MODEL_MAP = {
            "hubert": HubertModel,
            "wav2vec2": Wav2Vec2Model,
            "wavlm": WavLMModel,
            "whisper": WhisperModel,
        }

        for key, model_class in MODEL_MAP.items():
            if key in self.args.audio_model:
                audio_encoder = model_class.from_pretrained(self.args.audio_model)
                if key == 'whisper':
                    audio_encoder = audio_encoder.encoder
                break
                     
        else:
            available_models = ", ".join(MODEL_MAP.keys())
            raise ValueError(f"Unknown audio model '{self.args.audio_model}'. Available models are: {available_models}.")
        
        for param_name, param in audio_encoder.named_parameters():
            # if "feature_projection" in param_name or "feature_extractor" in param_name or 'conv' in param_name or "masked_spec_embed" in param_name:
            if 'feature_projection' in param_name or 'feature_extractor' in param_name:
                param.requires_grad = False
            else:
                param.requires_grad = True

        #region frozen or adapter
        # update 2023.11.03 
        # the same setting as text model
        if 'a' in frozen_mode:
            for param in audio_encoder.parameters():
                param.requires_grad = False
        
        elif 'a' in adapter_mode:
            for i in range(0, audio_encoder.config.num_hidden_layers):
                hidden_size = audio_encoder.config.hidden_size
                adapter = Adapter(hidden_size, hidden_size*4)
                setattr(audio_encoder.encoder.layers[i].attention, 'adapter', adapter)
                # setattr(audio_encoder.encoder.layers[i].feed_forward, 'adapter', adapter)

            for param_name, param in audio_encoder.named_parameters():
                if 'adapter' in param_name or 'LayerNorm' in param_name:
                    param.requires_grad = True
                else:
                    param.requires_grad = False                   
        #endregion
        
        return audio_encoder


    def kd_loss(self, teacher_logits, student_logits, teacher_feature, student_feature, labels):
        # set loss functions
        kl_loss = KLDivergenceLoss()
        mmd = MMD(kernel=self.args.kernel, kernel_params=self.args.kernel_params)
        info_nce = InfoNCE_KDLoss(info_temp=self.args.info_temp)
        
        # knowledge distillation method, select NST_KD, CRD_KD or KL_KD
        if self.args.kd_mode == 'nst': # NST_KD
            kd_loss = mmd.get_loss(teacher_feature, student_feature)
        elif self.args.kd_mode == 'crd': # CRD_KD
            kd_loss = info_nce.get_loss(teacher_feature, student_feature) 
        elif self.args.kd_mode == 'kl': # KL_KD
            kd_loss = kl_loss.get_loss(student_logits, teacher_logits)
        elif self.args.kd_mode == 'nst_kl': # NST_KD + KL_KD
            kd_loss = mmd.get_loss(teacher_feature, student_feature) + kl_loss.get_loss(student_logits, teacher_logits)
        elif self.args.kd_mode == 'crd_kl': # CRD_KD + KL_KD
            kd_loss = info_nce.get_loss(teacher_feature, student_feature) + kl_loss.get_loss(student_logits, teacher_logits)
        else:  # no knowledge distillation
            kd_loss = 0.0
        
        return kd_loss
    

    def get_confidence_score_train(self, logits, labels):
        # change the shape of labels to one-hot vector
        labels_one_hot = torch.zeros_like(logits)
        labels_one_hot.scatter_(1, labels.unsqueeze(1), 1)
        # calculate the confidence score
        conf = (F.softmax(logits, dim=-1) * labels_one_hot).sum(dim=-1)
        return conf
    

    def get_confidence_score_eval(self, logits):
        conf = F.softmax(logits, dim=-1).max(dim=-1)[0]
        return conf


    def forward(self, audio, input_ids, attention_mask, queries, text=None, labels=None):
        return_dict = {}
        ce_loss = CrossEntropyLoss()

        t_output = self.text_encoder(input_ids=input_ids, attention_mask=attention_mask, output_hidden_states=True)
        a_output = self.audio_encoder(audio, output_hidden_states=True)

        batch_size = input_ids.shape[0]
        text_seq_len, text_hidden_size = t_output.hidden_states[0].shape[1], t_output.hidden_states[0].shape[2]
        audio_seq_len, audio_hidden_size = a_output.hidden_states[0].shape[1], a_output.hidden_states[0].shape[2]

        assert text_hidden_size == audio_hidden_size, "text_hidden_size != audio_hidden_size"

        # shared_neck = self.bottleneck_embedding(self.bottleneck.expand(batch_size, self.bottleneck_len))
        shared_neck = self.bottleneck_embedding(queries)  # (batch_size, bottleneck_len, hidden_size)
        
        a_confs = []
        for i in range(self.fusion_layers, 0, -1):
            shared_neck = self.query_encoder[-i](shared_neck.permute(1,0,2))  # (len, batch_size, hidden_size)
            resicual_neck = shared_neck  # (len, batch_size, hidden_size)
            output_t = t_output.hidden_states[-i].permute(1,0,2)  # (len, batch_size, hidden_size)
            output_a = a_output.hidden_states[-i].permute(1,0,2)  # (len, batch_size, hidden_size)

            # cross modal encoder
            qt = self.cme_qt[-i](shared_neck, output_t, output_t)  # (len, batch_size, hidden_size)
            qa = self.cme_qa[-i](shared_neck, output_a, output_a)  # (len, batch_size, hidden_size)
            
            # fusion
            a_conf = self.conf_net(torch.mean(qt, dim=0))
            a_conf_expanded = a_conf.expand(shared_neck.shape[0], -1, shared_neck.shape[2])  # (len, batch_size, hidden_size)
            # if a_confidence is too low , we dont add it to the fusion feature
            a_conf_expanded = torch.where(a_conf_expanded > self.tau, a_conf_expanded, torch.zeros_like(a_conf_expanded))
            
            shared_neck = a_conf_expanded*qt + (1-a_conf_expanded)*qa  # (len, batch_size, hidden_size)
            shared_neck = shared_neck + resicual_neck  # (len, batch_size, hidden_size)
            shared_neck = shared_neck.permute(1,0,2)  # (batch_size, len, hidden_size)
            a_conf_expanded = a_conf_expanded.permute(1,0,2)  # (batch_size, len, hidden_size) 
            # shared_neck = self.a_weight*z_fsn_a + self.t_weight*z_fsn_t
            # save a_conf with the layer number
            a_confs.append(a_conf.squeeze().detach().cpu().numpy())
            
        fusion_pred = torch.mean(shared_neck, dim=1) # (batch_size, bottleneck_len, output_dim)
        fusion_logits = self.pred_head_union(fusion_pred) # (batch_size, bottleneck_len, output_dim)

        output_a = a_output.last_hidden_state
        a_pred = torch.mean(output_a, dim=1)
        a_logits = self.pred_head_a(a_pred)
        
        output_t = t_output.last_hidden_state
        t_pred = torch.mean(output_t, dim=1)
        t_logits = self.pred_head_t(t_pred)

        if labels is not None:
            t_ce = ce_loss.get_loss(t_logits, labels)
            a_ce = ce_loss.get_loss(a_logits, labels)
            fusion_ce = ce_loss.get_loss(fusion_logits, labels)
            loss =self.t_loss_weight*t_ce + self.a_loss_weight*a_ce + self.f_loss_weight*fusion_ce 
            # loss = t_ce
            logits = fusion_logits
            # logits = t_logits

        if self.train_mode:
            return_dict["t_ce_loss"] = t_ce
            return_dict["a_ce_loss"] = a_ce
            return_dict["fusion_ce_loss"] = fusion_ce
        else:
            return_dict['a_logits'] = a_logits
            return_dict['t_logits'] = t_logits
            return_dict['t_feature'] = t_pred
            return_dict['a_feature'] = a_pred
            return_dict['fusion_feature'] = fusion_pred
            return_dict['a_confs'] = a_confs

        return {"loss": loss, 
                "logits": logits, 
                "return_dict": return_dict}


class DQ_old_mask_text(DQ_old):
    def __init__(self, args):
        super(DQ_old_mask_text, self).__init__(args)


    def forward(self, audio, input_ids, attention_mask, queries, text=None, labels=None):
        return_dict = {}
        ce_loss = CrossEntropyLoss()

        t_output = self.text_encoder(input_ids=input_ids, attention_mask=attention_mask, output_hidden_states=True)
        a_output = self.audio_encoder(audio, output_hidden_states=True)

        batch_size = input_ids.shape[0]
        text_seq_len, text_hidden_size = t_output.hidden_states[0].shape[1], t_output.hidden_states[0].shape[2]
        audio_seq_len, audio_hidden_size = a_output.hidden_states[0].shape[1], a_output.hidden_states[0].shape[2]

        assert text_hidden_size == audio_hidden_size, "text_hidden_size != audio_hidden_size"

        # shared_neck = self.bottleneck_embedding(self.bottleneck.expand(batch_size, self.bottleneck_len))
        shared_neck = self.bottleneck_embedding(queries)  # (batch_size, bottleneck_len, hidden_size)
        
        a_confs = []
        for i in range(self.fusion_layers, 0, -1):
            shared_neck = self.query_encoder[-i](shared_neck.permute(1,0,2))  # (len, batch_size, hidden_size)
            resicual_neck = shared_neck  # (len, batch_size, hidden_size)
            output_t = t_output.hidden_states[-i].permute(1,0,2)  # (len, batch_size, hidden_size)
            output_a = a_output.hidden_states[-i].permute(1,0,2)  # (len, batch_size, hidden_size)
            output_t = 0 * output_t

            # cross modal encoder
            qt = self.cme_qt[-i](shared_neck, output_t, output_t)  # (len, batch_size, hidden_size)
            qa = self.cme_qa[-i](shared_neck, output_a, output_a)  # (len, batch_size, hidden_size)
            
            # fusion
            a_conf = self.conf_net(torch.mean(qt, dim=0))
            a_conf_expanded = a_conf.expand(shared_neck.shape[0], -1, shared_neck.shape[2])  # (len, batch_size, hidden_size)
            # if a_confidence is too low , we dont add it to the fusion feature
            a_conf_expanded = torch.where(a_conf_expanded > self.tau, a_conf_expanded, torch.zeros_like(a_conf_expanded))
            
            shared_neck = a_conf_expanded*qt + (1-a_conf_expanded)*qa  # (len, batch_size, hidden_size)
            shared_neck = shared_neck + resicual_neck  # (len, batch_size, hidden_size)
            shared_neck = shared_neck.permute(1,0,2)  # (batch_size, len, hidden_size)
            a_conf_expanded = a_conf_expanded.permute(1,0,2)  # (batch_size, len, hidden_size) 
            # shared_neck = self.a_weight*z_fsn_a + self.t_weight*z_fsn_t
            # save a_conf with the layer number
            a_confs.append(a_conf.squeeze().detach().cpu().numpy())
            
        fusion_pred = torch.mean(shared_neck, dim=1) # (batch_size, bottleneck_len, output_dim)
        fusion_logits = self.pred_head_union(fusion_pred) # (batch_size, bottleneck_len, output_dim)

        output_a = a_output.last_hidden_state
        a_pred = torch.mean(output_a, dim=1)
        a_logits = self.pred_head_a(a_pred)
        
        output_t = t_output.last_hidden_state
        t_pred = torch.mean(output_t, dim=1)
        t_logits = self.pred_head_t(t_pred)

        if labels is not None:
            t_ce = ce_loss.get_loss(t_logits, labels)
            a_ce = ce_loss.get_loss(a_logits, labels)
            fusion_ce = ce_loss.get_loss(fusion_logits, labels)
            loss =self.t_loss_weight*t_ce + self.a_loss_weight*a_ce + self.f_loss_weight*fusion_ce 
            # loss = t_ce
            logits = a_logits
            # logits = t_logits

        if self.train_mode:
            return_dict["t_ce_loss"] = t_ce
            return_dict["a_ce_loss"] = a_ce
            return_dict["fusion_ce_loss"] = fusion_ce
        else:
            return_dict['a_logits'] = a_logits
            return_dict['t_logits'] = t_logits
            return_dict['t_feature'] = t_pred
            return_dict['a_feature'] = a_pred
            return_dict['fusion_feature'] = fusion_pred
            return_dict['a_confs'] = a_confs

        return {"loss": loss, 
                "logits": logits, 
                "return_dict": return_dict}



class DQ_old_mask_audio(DQ_old):
    def __init__(self, args):
        super(DQ_old_mask_audio, self).__init__(args)


    def forward(self, audio, input_ids, attention_mask, queries, text=None, labels=None):
        return_dict = {}
        ce_loss = CrossEntropyLoss()

        t_output = self.text_encoder(input_ids=input_ids, attention_mask=attention_mask, output_hidden_states=True)
        a_output = self.audio_encoder(audio, output_hidden_states=True)

        batch_size = input_ids.shape[0]
        text_seq_len, text_hidden_size = t_output.hidden_states[0].shape[1], t_output.hidden_states[0].shape[2]
        audio_seq_len, audio_hidden_size = a_output.hidden_states[0].shape[1], a_output.hidden_states[0].shape[2]

        assert text_hidden_size == audio_hidden_size, "text_hidden_size != audio_hidden_size"

        # shared_neck = self.bottleneck_embedding(self.bottleneck.expand(batch_size, self.bottleneck_len))
        shared_neck = self.bottleneck_embedding(queries)  # (batch_size, bottleneck_len, hidden_size)
        
        a_confs = []
        for i in range(self.fusion_layers, 0, -1):
            shared_neck = self.query_encoder[-i](shared_neck.permute(1,0,2))  # (len, batch_size, hidden_size)
            resicual_neck = shared_neck  # (len, batch_size, hidden_size)
            output_t = t_output.hidden_states[-i].permute(1,0,2)  # (len, batch_size, hidden_size)
            output_a = a_output.hidden_states[-i].permute(1,0,2)  # (len, batch_size, hidden_size)
            output_a = 0 * output_a

            # cross modal encoder
            qt = self.cme_qt[-i](shared_neck, output_t, output_t)  # (len, batch_size, hidden_size)
            qa = self.cme_qa[-i](shared_neck, output_a, output_a)  # (len, batch_size, hidden_size)
            
            # fusion
            a_conf = self.conf_net(torch.mean(qt, dim=0))
            a_conf_expanded = a_conf.expand(shared_neck.shape[0], -1, shared_neck.shape[2])  # (len, batch_size, hidden_size)
            # if a_confidence is too low , we dont add it to the fusion feature
            a_conf_expanded = torch.where(a_conf_expanded > self.tau, a_conf_expanded, torch.zeros_like(a_conf_expanded))
            
            shared_neck = a_conf_expanded*qt + (1-a_conf_expanded)*qa  # (len, batch_size, hidden_size)
            shared_neck = shared_neck + resicual_neck  # (len, batch_size, hidden_size)
            shared_neck = shared_neck.permute(1,0,2)  # (batch_size, len, hidden_size)
            a_conf_expanded = a_conf_expanded.permute(1,0,2)  # (batch_size, len, hidden_size) 
            # shared_neck = self.a_weight*z_fsn_a + self.t_weight*z_fsn_t
            # save a_conf with the layer number
            a_confs.append(a_conf.squeeze().detach().cpu().numpy())
            
        fusion_pred = torch.mean(shared_neck, dim=1) # (batch_size, bottleneck_len, output_dim)
        fusion_logits = self.pred_head_union(fusion_pred) # (batch_size, bottleneck_len, output_dim)

        output_a = a_output.last_hidden_state
        a_pred = torch.mean(output_a, dim=1)
        a_logits = self.pred_head_a(a_pred)
        
        output_t = t_output.last_hidden_state
        t_pred = torch.mean(output_t, dim=1)
        t_logits = self.pred_head_t(t_pred)

        if labels is not None:
            t_ce = ce_loss.get_loss(t_logits, labels)
            a_ce = ce_loss.get_loss(a_logits, labels)
            fusion_ce = ce_loss.get_loss(fusion_logits, labels)
            loss =self.t_loss_weight*t_ce + self.a_loss_weight*a_ce + self.f_loss_weight*fusion_ce 
            # loss = t_ce
            logits = t_logits
            # logits = t_logits

        if self.train_mode:
            return_dict["t_ce_loss"] = t_ce
            return_dict["a_ce_loss"] = a_ce
            return_dict["fusion_ce_loss"] = fusion_ce
        else:
            return_dict['a_logits'] = a_logits
            return_dict['t_logits'] = t_logits
            return_dict['t_feature'] = t_pred
            return_dict['a_feature'] = a_pred
            return_dict['fusion_feature'] = fusion_pred
            return_dict['a_confs'] = a_confs

        return {"loss": loss, 
                "logits": logits, 
                "return_dict": return_dict}