
import torch
import torch.nn as nn
import torch.utils.checkpoint
from torch.nn import TransformerEncoderLayer
from utils.loss import CrossEntropyLoss
from modules.transformer import TransformerEncoderLayer as Cross_model_encoder
from utils.helper import PredictLayer
from .model import MyModel

class DQ_wo_AMW(MyModel):
    def __init__(self,args):
        super(DQ_wo_AMW, self).__init__(args)
        self.num_heads = args.num_heads
        self.fusion_layers = args.fusion_layers
        self.relu_dropout = args.relu_dropout
        self.res_dropout = args.res_dropout
        self.adapter_dim = args.adapter_dim
        self.bottleneck_len = args.bottleneck_len
    
        self.text_encoder = self.get_text_encoder()
        self.audio_encoder = self.get_audio_encoder()
        self.bottleneck_embedding = nn.Embedding(self.bottleneck_len, self.intermediate_dim)
        self.query_encoder = nn.ModuleList([
            TransformerEncoderLayer(d_model=self.intermediate_dim, nhead=self.num_heads, batch_first=False, dropout=self.dropout, activation='gelu')
            for _ in range(self.fusion_layers)])
        # cross modal encoder
        self.cme_qt = nn.ModuleList([
            Cross_model_encoder(embed_dim=self.intermediate_dim, num_heads=self.num_heads, attn_dropout=self.dropout, relu_dropout=self.relu_dropout, res_dropout=self.res_dropout)
            for _ in range(self.fusion_layers)])
        self.cme_qa = nn.ModuleList([
            Cross_model_encoder(embed_dim=self.intermediate_dim, num_heads=self.num_heads, attn_dropout=self.dropout, relu_dropout=self.relu_dropout, res_dropout=self.res_dropout)
            for _ in range(self.fusion_layers)])

    
        # prediciton head
        self.pred_head_t = PredictLayer(self.intermediate_dim, self.output_dim, dropout=self.dropout, residual=True)
        self.pred_head_a = PredictLayer(self.intermediate_dim, self.output_dim, dropout=self.dropout, residual=True)
        self.pred_head_union = PredictLayer(self.intermediate_dim, self.output_dim, dropout=self.dropout, residual=True)

    def forward(self, audio_value, audio_attention_mask, input_ids, attention_mask, queries, text=None, labels=None):
        return_dict = {}
        ce_loss = CrossEntropyLoss()

        t_output = self.text_encoder(input_ids=input_ids, attention_mask=attention_mask, output_hidden_states=True)
        a_output = self.audio_encoder(audio_value, output_hidden_states=True)

        text_seq_len, text_hidden_size = t_output.hidden_states[0].shape[1], t_output.hidden_states[0].shape[2]
        audio_seq_len, audio_hidden_size = a_output.hidden_states[0].shape[1], a_output.hidden_states[0].shape[2]

        assert text_hidden_size == audio_hidden_size, "text_hidden_size != audio_hidden_size"

        # shared_neck = self.bottleneck_embedding(self.bottleneck.expand(batch_size, self.bottleneck_len))
        shared_neck = self.bottleneck_embedding(queries)  # (batch_size, bottleneck_len, hidden_size)
        
        for i in range(self.fusion_layers, 0, -1):
            shared_neck = self.query_encoder[-i](shared_neck.permute(1,0,2))  # (len, batch_size, hidden_size)
            resicual_neck = shared_neck  # (len, batch_size, hidden_size)
            output_t = t_output.hidden_states[-i].permute(1,0,2)  # (len, batch_size, hidden_size)
            output_a = a_output.hidden_states[-i].permute(1,0,2)  # (len, batch_size, hidden_size)

            # cross modal encoder
            qt = self.cme_qt[-i](shared_neck, output_t, output_t)  # (len, batch_size, hidden_size)
            qa = self.cme_qa[-i](shared_neck, output_a, output_a)  # (len, batch_size, hidden_size)
            
            # fusion
            shared_neck = 0.5*qt + 0.5*qa  # (len, batch_size, hidden_size)
            shared_neck = shared_neck + resicual_neck  # (len, batch_size, hidden_size)
            shared_neck = shared_neck.permute(1,0,2)  # (batch_size, len, hidden_size)
            
        fusion_pred = torch.mean(shared_neck, dim=1) # (batch_size, bottleneck_len, output_dim)
        fusion_logits = self.pred_head_union(fusion_pred) # (batch_size, bottleneck_len, output_dim)

        output_a = a_output.last_hidden_state
        a_pred = torch.mean(output_a, dim=1)
        a_logits = self.pred_head_a(a_pred)
        
        output_t = t_output.last_hidden_state
        t_pred = torch.mean(output_t, dim=1)
        t_logits = self.pred_head_t(t_pred)

        if labels is not None:
            t_loss = self.calculate_task_loss(t_logits, labels)
            a_loss = self.calculate_task_loss(a_logits, labels)
            f_loss = self.calculate_task_loss(fusion_logits, labels)
            loss = t_loss + a_loss + f_loss
            # t_ce = ce_loss.get_loss(t_logits, labels)
            # a_ce = ce_loss.get_loss(a_logits, labels)
            # fusion_ce = ce_loss.get_loss(fusion_logits, labels)
            # loss = t_ce + a_ce + fusion_ce 
            logits = fusion_logits
        else:
            loss = None
            logits = fusion_logits

        if self.train_mode:
            return_dict["t_loss"] = t_loss
            return_dict["a_loss"] = a_loss
            return_dict["f_loss"] = f_loss
        else:
            return_dict['a_logits'] = a_logits
            return_dict['t_logits'] = t_logits
            return_dict['t_feature'] = t_pred
            return_dict['a_feature'] = a_pred
            return_dict['fusion_feature'] = fusion_pred

        return {"loss": loss, 
                "logits": logits, 
                "return_dict": return_dict}