import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.checkpoint
from utils.loss import CrossEntropyLoss, KLDivergenceLoss
from modules.my_transformer import MyTransformerEncoderLayer
from .model import MyModel
from utils.helper import PredictLayer
from transformers import BertModel, BertConfig

class Semantic_encoder(nn.Module):
    def __init__(self, n_heads, n_layers, d_model, dropout):
        super(Semantic_encoder, self).__init__()
        self.semantic_encoder = nn.ModuleList([
            MyTransformerEncoderLayer(d_model=d_model, nhead=n_heads, dropout=dropout, batch_first=True)
            for _ in range(n_layers)])
    
    def forward(self, x, need_weights=False):
        """
            Forward pass of the semantic encoder.

            Args:
                x (torch.Tensor): Input tensor.
                need_weights (bool): Flag indicating whether to return attention weights.

            Returns:
                torch.Tensor or (torch.Tensor, list): Encoded representation of input tensor.
                    If need_weights is True, also returns a list of attention weights.
        """
        if need_weights:
            attention_weights = []
            for i in range(len(self.semantic_encoder)):
                x, attn_weights = self.semantic_encoder[i].forward(x, need_weights=True)
                attention_weights.append(attn_weights.detach().cpu().numpy())
            return x, attention_weights
        else:
            for i in range(len(self.semantic_encoder)):
                x = self.semantic_encoder[i].forward(x)
            return x


class Single_stream(MyModel):
    def __init__(self,args):
        super(Single_stream, self).__init__(args)
        self.d_model = args.intermediate_dim
        self.n_head = args.num_heads
        self.n_layer = args.num_layers
        self.dropout = args.dropout
        self.enable_modality_embedding = args.enable_modality_embedding

        self.text_encoder = self.get_text_encoder(adapter_mode=False, frozen_mode=False)
        self.text_dim = self.text_encoder.config.hidden_size
        self.proj_text = nn.Conv1d(in_channels=self.text_dim, out_channels=self.d_model, kernel_size=1)

        self.audio_encoder = self.get_audio_encoder(adapter_mode=False, frozen_mode=False)
        self.audio_dim = self.audio_encoder.config.hidden_size
        self.proj_audio = nn.Conv1d(in_channels=self.audio_dim, out_channels=self.d_model, kernel_size=4, stride=2, padding=2)

        # Define embedding layer 1: text, 2: audio, 3: video
        self.modality_embedding_func = nn.Embedding(num_embeddings=4, embedding_dim=self.d_model, padding_idx=0) 
        self.shared_semantic_encoder = Semantic_encoder(self.n_head, self.n_layer, self.d_model, self.dropout)

        self.pred_head = PredictLayer(self.d_model, self.output_dim, dropout=self.dropout, residual=True)
        

    def modality_embedding(self, 
                           seq_len : int = None, 
                           modality_type : str = None):
        modality_type_mapping = {'text': 1, 'audio': 2, 'vision': 3}
        assert  modality_type in modality_type_mapping, "modality_type should be one of text, audio, vision"
        modality_type_id = modality_type_mapping.get(modality_type, 0)
        token_type_id = torch.full((seq_len,), modality_type_id, dtype=torch.long).to('cuda')
        modality_embedding = self.modality_embedding_func(token_type_id)
        return modality_embedding


    def forward(self, audio_value=None, audio_attention_mask=None, input_ids=None, attention_mask=None, queries=None, text=None, labels=None):
        return_dict = {}
        #region get input embeddings
        if audio_value is not None:
            a_features = self.audio_encoder(audio_value, output_hidden_states=True)
            a_input_embedding = a_features.last_hidden_state # (batch_size, seq_len, audio_dim)
            a_input_embedding = self.proj_audio(a_input_embedding.permute(0, 2, 1)).permute(0, 2, 1)
            if self.enable_modality_embedding:
                audio_token_type_idx = 2
                a_token_embeds = self.modality_embedding_func(torch.full((a_input_embedding.size(1),), audio_token_type_idx, dtype=torch.long, device=a_input_embedding.device))
                a_embedding = a_input_embedding + a_token_embeds
                # a_embedding = a_input_embedding + self.modality_embedding(seq_len=a_input_embedding.size(1), modality_type='audio')
            else:
                a_embedding = a_input_embedding

        if input_ids is not None:
            t_features = self.text_encoder(input_ids=input_ids, attention_mask=attention_mask, output_hidden_states=True)
            t_input_embedding = t_features.last_hidden_state # (batch_size, seq_len, text_dim)
            t_input_embedding = self.proj_text(t_input_embedding.permute(0,2,1)).permute(0,2,1)
            if self.enable_modality_embedding:
                text_token_type_idx = 1
                t_token_embeds = self.modality_embedding_func(torch.full((t_input_embedding.size(1),), text_token_type_idx, dtype=torch.long, device=t_input_embedding.device))
                t_embedding = t_input_embedding + t_token_embeds
                # t_embedding = t_input_embedding + self.modality_embedding(seq_len=t_input_embedding.size(1), modality_type='text')
            else:
                t_embedding = t_input_embedding
        #endregion
                
        #region get multimodal embedding
        if a_embedding is not None and t_embedding is not None:
            multi_embedding = torch.cat([t_embedding, a_embedding], dim=1)
        elif a_embedding is not None:
            multi_embedding = a_embedding
        elif t_embedding is not None:
            multi_embedding = t_embedding
        else:
            raise ValueError("Both audio and text inputs cannot be None.")
        
        multi_embedding = self.shared_semantic_encoder(multi_embedding, need_weights=False)
        #endregion

        # average pooling
        fusion_pred = torch.mean(multi_embedding, dim=1)
        fusion_logits = self.pred_head(fusion_pred)
        
        if labels is not None:
            loss = self.calculate_task_loss(fusion_logits, labels)
            logits = fusion_logits
        else:
            loss = None
            logits = fusion_logits

        return {"loss": loss, 
                "logits": logits, 
                "return_dict": return_dict}


class Single_stream_mask_text(Single_stream):
    def __init__(self, args):
        super(Single_stream_mask_text, self).__init__(args)

    
    def forward(self, audio_value=None, audio_attention_mask=None, input_ids=None, attention_mask=None, queries=None, text=None, labels=None):
        return_dict = {}
        #region get input embeddings
        if audio_value is not None:
            a_features = self.audio_encoder(audio_value, output_hidden_states=True)
            a_input_embedding = a_features.last_hidden_state # (batch_size, seq_len, audio_dim)
            a_input_embedding = self.proj_audio(a_input_embedding.permute(0, 2, 1)).permute(0, 2, 1)
            if self.enable_modality_embedding:
                audio_token_type_idx = 2
                a_token_embeds = self.modality_embedding_func(torch.full((a_input_embedding.size(1),), audio_token_type_idx, dtype=torch.long, device=a_input_embedding.device))
                a_embedding = a_input_embedding + a_token_embeds
                # a_embedding = a_input_embedding + self.modality_embedding(seq_len=a_input_embedding.size(1), modality_type='audio')
            else:
                a_embedding = a_input_embedding

        if input_ids is not None:
            t_features = self.text_encoder(input_ids=input_ids, attention_mask=attention_mask, output_hidden_states=True)
            t_input_embedding = t_features.last_hidden_state # (batch_size, seq_len, text_dim)
            t_input_embedding = self.proj_text(t_input_embedding.permute(0,2,1)).permute(0,2,1)
            if self.enable_modality_embedding:
                text_token_type_idx = 1
                t_token_embeds = self.modality_embedding_func(torch.full((t_input_embedding.size(1),), text_token_type_idx, dtype=torch.long, device=t_input_embedding.device))
                t_embedding = t_input_embedding + t_token_embeds
                # t_embedding = t_input_embedding + self.modality_embedding(seq_len=t_input_embedding.size(1), modality_type='text')
            else:
                t_embedding = t_input_embedding
        #endregion
                
        #region get multimodal embedding
        if a_embedding is not None and t_embedding is not None:
            # multi_embedding = torch.cat([0 * t_embedding, a_embedding], dim=1)
            multi_embedding = a_embedding
        elif a_embedding is not None:
            multi_embedding = a_embedding
        elif t_embedding is not None:
            multi_embedding = t_embedding
        else:
            raise ValueError("Both audio and text inputs cannot be None.")
        
        multi_embedding = self.shared_semantic_encoder(multi_embedding, need_weights=False)
        #endregion

        # average pooling
        fusion_pred = torch.mean(multi_embedding, dim=1)
        fusion_logits = self.pred_head(fusion_pred)
        
        if labels is not None:
            loss = self.calculate_task_loss(fusion_logits, labels)
            logits = fusion_logits

        else:
            loss = None
            logits = fusion_logits

        return {"loss": loss, 
                "logits": logits, 
                "return_dict": return_dict}
    

class Single_stream_mask_audio(Single_stream):
    def __init__(self, args):
        super(Single_stream_mask_audio, self).__init__(args)

    def forward(self, audio_value=None, audio_attention_mask=None, input_ids=None, attention_mask=None, queries=None, text=None, labels=None):
        return_dict = {}
        #region get input embeddings
        if audio_value is not None:
            a_features = self.audio_encoder(audio_value, output_hidden_states=True)
            a_input_embedding = a_features.last_hidden_state # (batch_size, seq_len, audio_dim)
            a_input_embedding = self.proj_audio(a_input_embedding.permute(0, 2, 1)).permute(0, 2, 1)
            if self.enable_modality_embedding:
                audio_token_type_idx = 2
                a_token_embeds = self.modality_embedding_func(torch.full((a_input_embedding.size(1),), audio_token_type_idx, dtype=torch.long, device=a_input_embedding.device))
                a_embedding = a_input_embedding + a_token_embeds
                # a_embedding = a_input_embedding + self.modality_embedding(seq_len=a_input_embedding.size(1), modality_type='audio')
            else:
                a_embedding = a_input_embedding

        if input_ids is not None:
            t_features = self.text_encoder(input_ids=input_ids, attention_mask=attention_mask, output_hidden_states=True)
            t_input_embedding = t_features.last_hidden_state # (batch_size, seq_len, text_dim)
            t_input_embedding = self.proj_text(t_input_embedding.permute(0,2,1)).permute(0,2,1)
            if self.enable_modality_embedding:
                text_token_type_idx = 1
                t_token_embeds = self.modality_embedding_func(torch.full((t_input_embedding.size(1),), text_token_type_idx, dtype=torch.long, device=t_input_embedding.device))
                t_embedding = t_input_embedding + t_token_embeds
                # t_embedding = t_input_embedding + self.modality_embedding(seq_len=t_input_embedding.size(1), modality_type='text')
            else:
                t_embedding = t_input_embedding
        #endregion
                
        #region  get multimodal embedding
        if a_embedding is not None and t_embedding is not None:
            multi_embedding = torch.cat([t_embedding, 0 * a_embedding], dim=1)
            # multi_embedding = t_embedding
        elif a_embedding is not None:
            multi_embedding = a_embedding
        elif t_embedding is not None:
            multi_embedding = t_embedding
        else:
            raise ValueError("Both audio and text inputs cannot be None.")
        
        multi_embedding = self.shared_semantic_encoder(multi_embedding, need_weights=False)
        #endregion

        # average pooling
        fusion_pred = torch.mean(multi_embedding, dim=1)
        fusion_logits = self.pred_head(fusion_pred)
        
        if labels is not None:
            loss = self.calculate_task_loss(fusion_logits, labels)
            logits = fusion_logits

        else:
            loss = None
            logits = fusion_logits

        return {"loss": loss, 
                "logits": logits, 
                "return_dict": return_dict}